/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2007 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/*******************************************************
* non-smooth coefficients elliptic PDE
* 
* -div(a(x) grad u) = f  x \in \Omega
* u = u0  x \in \partial \Omega
*
* a(x): piesewise constant  a(x) > \alpha > 0
* a(x) = 1 in first and third quadrants of x-y
* a(x) = R in second and fourth quadrants of x-y
* ture solution is func_g
* the function is modified from: R. B. Kellogg, 
* on the poisson equation with intersecting interface, 
* Applicable Analysis 4, 101-129
*
* k(x, y) is the function defined by Kellogg
* then u(x, y, z) = k(x,y) + A_CONS * z + B_CONS :=func_g
*
* We should say that \rho, \sigma, R, \gamma must
* satisfy some relations. [Kellog]
* By [Kellogg], f is constant and f is equal to 0.0.
*
* $Id: non-smooth.c,v 1.22 2007/09/27 10:56:56 zlb Exp $
*******************************************************/

#include "phg.h"
#include <math.h>

#define TOL 2.0
#define OUTPUT_MESH 1       /* 1:output   0: not output */
#define refine_max 0.5      /* max strategy: theta */
#define DOFTYPE DOF_DEFAULT

/* parameters used by function func_g and func_ax*/
#define PI      3.14159265358
#define R       161.44763879
#define gamma   0.1
#define rho     0.78539816339    /* PI / 4.0 */
#define sigma   -14.92256510455
#define A_CONS  4.0
#define B_CONS  10.0

static void   /* rhs */
func_f(FLOAT x, FLOAT y, FLOAT z, FLOAT *value)
{
    *value = 0.0;
}

static void  /* boundary function */
func_g(FLOAT x, FLOAT y, FLOAT z, FLOAT *value)
{
    FLOAT r,theta = 0.0, tmp, tmp1, rtmp;
    
    tmp = x * x + y * y;
    r = Sqrt(tmp);
    rtmp = Pow(r, gamma);
    
    if (x >= 0.0 && y >= 0.0 && r > 0.0){ /* first quadrant:x-y */
        tmp = y / r;
	theta = Asin(tmp);
	tmp = (PI * 0.5 - sigma) * gamma;
	tmp1 = (theta - PI * 0.5 + rho) * gamma;
	*value = Cos(tmp) * Cos(tmp1) * rtmp + A_CONS * z + B_CONS;
    }
    else if (x <= 0.0 && y >= 0.0 && r > 0.0){ /* second quadrant:x-y */
        tmp = y / r;
	theta = PI - Asin(tmp);
	tmp = rho * gamma;
	tmp1 = (theta - PI + sigma) * gamma;
	*value = Cos(tmp) * Cos(tmp1) * rtmp + A_CONS * z + B_CONS;
    }
    else if (x <= 0.0 && y <= 0.0 && r > 0.0){ /* third quadrant:x-y */
        tmp = y / r;
	theta = PI - Asin(tmp);
	tmp = sigma * gamma;
	tmp1 = (theta - PI - rho) * gamma;
	*value = Cos(tmp) * Cos(tmp1) * rtmp + A_CONS * z + B_CONS;
    }
    else if (x >= 0.0 && y <= 0.0 && r > 0.0){ /* fourth quadrant:x-y */
        tmp = y / r;
	theta = 2.0 * PI + Asin(tmp);
	tmp = (PI * 0.5 - rho) * gamma;
	tmp1 = (theta - PI * 1.5 - sigma) * gamma;
	*value = Cos(tmp) * Cos(tmp1) * rtmp + A_CONS * z + B_CONS;
    }
    else if (r == 0.0){
        *value = A_CONS * z + B_CONS;
    }
}

static void /* coefficent a(x)*/
func_ax(FLOAT x, FLOAT y, FLOAT z, FLOAT *value)
{
    if ((x > 0.0 && y > 0.0) || (x < 0.0 && y < 0.0)){
        *value = R;
    }
    else if ((x < 0.0 && y > 0.0) || ( x > 0.0 && y < 0.0)){
        *value = 1.0;
    }
}

static void /* coefficent a(x), 3 dim */
func_ax_3(FLOAT x, FLOAT y, FLOAT z, FLOAT *value)
{
    if ((x > 0.0 && y > 0.0) || (x < 0.0 && y < 0.0)){
        *(value++) = R;
        *(value++) = R;
        *(value++) = R;
    }
    else if ((x < 0.0 && y > 0.0) || ( x > 0.0 && y < 0.0)){
        *(value++) = 1.0;
        *(value++) = 1.0;
        *(value++) = 1.0;
    }
}

static FLOAT  /* calculate time */
cal_time(BOOLEAN flag)
{
    static FLOAT t0 = 0.0;
    double et, tt[3];

    phgGetTime(tt);
    et = tt[2] - t0;
    t0 = tt[2];

    if (flag){
        phgPrintf("  %0.4lfs", (double)et);
    }

    return et;
}

/* cal energy norm */
/* (\int a(x) \grad u dot \grad u dx */
static FLOAT 
phgDofNormEnergy(DOF *u, DOF *a)
{
    GRID *g = u->g;
    SIMPLEX *e;
    DOF *grad;
    FLOAT sum = 0.0;
    FLOAT tmp = 0.0;

    grad = phgDofGradient(u, NULL, NULL, "u grad");

    ForAllElements(g, e){
        tmp = *DofElementData(a, e->index);
	sum += tmp * phgQuadDofDotDof(e, grad, grad, -1);
    }
    phgDofFree(&grad);
    return Sqrt(sum);
}

/* build linear systme */
static void
build_linear_system(SOLVER *solver, DOF *u_h, DOF *f_h, DOF *a)
{
    int N = u_h->type->nbas;    /* number of basis functions in an element */
    GRID *g = u_h->g;
    SIMPLEX *e;
    int i, j;
    FLOAT A[N][N], B[N], buffer[N], *row, tmp;
    INT I[N];

    assert(u_h->dim == 1);

    ForAllElements(g, e) {
    /* compute \int a(x) \grad\phi_i \cdot \grad\phi_j making use of symmetry */
        tmp = *DofElementData(a, e->index);
        for (i = 0; i < N; i++) {
            I[i] = phgSolverMapE2L(solver, 0, e, i);
            for (j = 0; j <= i; j++)
                A[j][i] = A[i][j] =
                tmp * phgQuadGradBasDotGradBas(e, u_h, i, u_h, j, QUAD_DEFAULT);
	}

	/* loop on basis functions */
        for (i = 0; i < N; i++) {
            if (phgDofDirichletBC(u_h, e, i, func_g,
                  buffer, B + i, DOF_PROJ_NONE)) {
                row = buffer;
	    }
            else {
                /* right hand side: \int f * phi_i */
                phgQuadDofTimesBas(e, f_h, u_h, i, QUAD_DEFAULT, B + i);
                row = A[i];
	    }
            phgSolverAddMatrixEntries(solver, 1, I + i, N, I, row); 
	}
        phgSolverAddRHSEntries(solver, N, I, B);
    }
}

static void  /* solve Au = b */
SolveSolver(OEM_SOLVER *sol, DOF *u_h, DOF *f_h, DOF *ax)
{
    GRID *g = u_h->g;
    SOLVER *solver;
    int nits;
    phgPrintf("Number of DOF:           %d\n", DofGetDataCountGlobal(u_h));
    phgPrintf("Number of elements:      %d\n", g->nleaf_global);
    phgPrintf("Load imbalance:          %lg\n",  (double)g->lif);
                       
    phgPrintf("create solver:         ");
    cal_time(FALSE);
    solver = phgSolverCreate(sol, u_h, NULL);
    cal_time(TRUE);
    phgPrintf("\n");
                
    phgPrintf("build linear system:   ");
    cal_time(FALSE);
    build_linear_system(solver, u_h, f_h, ax);
    cal_time(TRUE);
    phgPrintf("\n");
              
    phgPrintf("solve solver:          ");
    cal_time(FALSE);
    nits = phgSolverSolve(solver, TRUE, u_h, NULL);
    cal_time(TRUE);
    phgPrintf("     [%d]\n", nits);
    phgSolverDestroy(&solver);

}

static FLOAT /* calulate space indicator */
estimate_space(DOF *u_h, DOF *f_h, DOF *a, DOF *error)
{
    GRID *g = u_h->g;
    SIMPLEX *e;
    NEIGHBOUR_DATA *aa;
    DOF *jump, *grad, *agrad, *div, *ax3, *temp;
    FLOAT sum, tmp;
    FLOAT h, ak, ae;
    int i;

    ax3 = phgDofNew(g, DOF_P0, 3, "ax3", func_ax_3);
    grad = phgDofGradient(u_h, NULL, NULL, "grad");
    agrad = phgDofMM(MAT_OP_N, MAT_OP_N, 3, 1, 3, 1.0, ax3, 1, grad, 0.0, NULL);
    temp = phgDofDivergence(agrad, NULL, NULL, "div");
    div = phgDofCopy(f_h, NULL, DOF_DGn[u_h->type->order], NULL);
    phgDofAXPY(1.0, temp, &div);
    phgDofFree(&temp);
    jump = phgQuadFaceJump(agrad, DOF_PROJ_DOT, "jump", -1);
    phgDofFree(&grad);

    phgDofSetDataByValue(error, 0.0);
    aa = phgGetNeighbourData(a);  
    ForAllElements(g, e){ /* neighbours are local */
        sum = 0.0;
        tmp = *DofElementData(a, e->index);
        h = phgGeomGetDiameter(g, e);
        sum = h * h / tmp * phgQuadDofDotDof(e, div, div, -1);
        for (i = 0; i < NFace; i++) {
            if (e->bound_type[i] & (DIRICHLET | NEUMANN))
                continue;    /* boundary face */
            ak = *phgNeighbourData(aa, e, i, ELEMENT, 0);
            ae = (tmp > ak) ? tmp : ak;
            h = phgGeomGetFaceDiameter(g, e, i);
            sum +=  (*DofFaceData(jump, e->faces[i])) * h * 0.5 / ae;
	}
        *DofElementData(error, e->index) = sum;
    }
    
    phgDofFree(&div);
    phgDofFree(&jump);
    phgDofFree(&ax3);
    phgDofFree(&agrad);
    phgReleaseNeighbourData(&aa);

    return phgDofNormInftyVec(error);

}

static void /* mark elements for refining: max strategy */
MarkElements(DOF *u_h, DOF *error, FLOAT PE_error, FLOAT etaspace)
{
    GRID *g = error->g;
    SIMPLEX *e;
    FLOAT tmpc;
    FLOAT tmp;
    long  countr, countc;
   
    /* clean marks */
    ForAllElements(g, e){
        e->mark = 0;
    }
   
    tmpc = refine_max * PE_error;
    countr = countc = 0;
    ForAllElements(g, e){
	tmp = *DofElementData(error, e->index);
        if (tmpc < tmp){
            e->mark = 1;
            countr++;
	}
    }
    
    countc = countr;
    
    phgPrintf("Elements marked for refining: %ld\n", countc);
    
}

int
main(int argc, char *argv[])
{
    GRID *g;
    DOF *u_h, *f_h, *ax, *error;
    DOF *u_t, *point = NULL;
    FLOAT etaspace, etamax, tmp, tp, tmp1;
    double tt1[3], tt2[3], tt3[3];
    size_t peakmem;
    static char *fn = "../test/non-smooth.dat";
    int flag = 0;
    
    phgInit(&argc, &argv);
    g = phgNewGrid(-1);
    if (!phgImport(g, fn, TRUE))
        phgError(1, "can't read file \"%s\".\n", fn);

    u_h = phgDofNew(g, DOFTYPE, 1, "u_h", DofNoAction);
    error = phgDofNew(g, DOF_P0, 1, "error indicator", DofNoAction);
    u_t = phgDofNew(g, DOFTYPE, 1, "u_t", func_g);
    f_h = phgDofNew(g, DOFTYPE, 1, "f_h", func_f);
    ax = phgDofNew(g, DOF_P0, 1, "a(x)", func_ax);

#if 0
    /* distribute grid to all processes */
    phgRefineAllElements(g, 7);
#endif
    phgPrintf("\nNumber of processes: %d\n\n", phgNProcs);
    phgGetTime(tt1);
    
    while(TRUE){
        phgGetTime(tt2);
	flag++;
	phgPrintf("\n>>>>>> %d\n", flag);
        if (phgBalanceGrid(g, 1.05, -1, NULL, 0.))
            phgPrintf("Repartition mesh, load imbalance: %lg\n",
                (double)g->lif);

	/* solve the linear system */
        SolveSolver(SOLVER_DEFAULT, u_h, f_h, ax);

	/* calculate error */
        phgPrintf("calculate errors:      ");
        cal_time(FALSE);
        etamax = estimate_space(u_h, f_h, ax, error);
	etaspace = phgDofNormL1Vec(error);
        cal_time(TRUE);
        phgPrintf("\nmax space indicator: %.8lg     total: %lg\n",
			(double)etamax, (double)etaspace);

	phgDofFree(&point);
	point = phgDofCopy(u_t, NULL, DOFTYPE, NULL);
	phgDofAXPY(-1.0, u_h, &point);

	/* calculate L2 norm */
	tmp = phgDofNormL2(u_t);
	tp = phgDofNormL2(point);
	tmp1 = phgDofNormL2(u_h);
        phgPrintf("True norm(L2 norm):                  %.8lg\n", (double)tmp);
        phgPrintf("numerical norm(L2 norm):             %.8lg\n", (double)tmp1);
        phgPrintf("True error(L2):                      %.8lg\n", (double)tp);
        phgPrintf("relative error(L2 norm):             %.8lg\n",
					(double)(tp / tmp));

	/* pointwise error */
	tp = phgDofNormInftyVec(u_t);
	tmp = phgDofNormInftyVec(point);
        phgPrintf("max point error(Infinity):           %.8lg\n", (double)tmp);
	phgPrintf("max relative error(Infinity):        %.8lg\n",
					(double)(tmp / tp));

	/* calculate energy norm */
	tmp = phgDofNormEnergy(u_t, ax);
	tp = phgDofNormEnergy(u_h, ax);
	tmp1 = phgDofNormEnergy(point, ax);
        phgPrintf("True norm(Energy norm):              %.8lg\n", (double)tmp);
        phgPrintf("numerical norm(Energy norm):         %.8lg\n", (double)tp);
        phgPrintf("True error(Energy):                  %.8lg\n", (double)tmp1);
        phgPrintf("relative error(Energy norm):         %.8lg\n",
					(double)(tmp1 / tmp));
	
        phgMemoryUsage(g, &peakmem);
        if (etaspace <= TOL || peakmem >= 1024 * (size_t)400 * 1024){
            phgGetTime(tt3);
            phgPrintf("time usage: %lfs\n", (double)(tt3[2] - tt2[2]));
            phgPrintf("memory usage: %lfMb\n\n",
					(double)peakmem / (1024. * 1024.));
            break;
	}
	/* mark elements for refining */
        MarkElements(u_h, error, etamax, etaspace);
        phgPrintf("refine elements:");
        cal_time(FALSE);
	
	/* refine elements */
        phgRefineMarkedElements(g);
        cal_time(TRUE);
        phgGetTime(tt3);
        phgPrintf("\ntime usage: %lfs\n", (double)(tt3[2] - tt2[2]));
        phgMemoryUsage(g, &peakmem);
        phgPrintf("memory usage: %lf Mb\n\n", (double)peakmem / (1024.*1024.));
    }

    phgPrintf("Total time: %lfs\n", (double)(tt3[2] - tt1[2]));
#if OUTPUT_MESH
    {
	DOF *grad;
	grad = phgDofGradient(u_h, NULL, NULL, "grad_u_h");
	phgPrintf("Final mesh written to \"%s\".\n",
			phgExportVTK(g, "non-smooth.vtk",
				u_h, grad, ax, error, u_t, point, NULL));
	phgDofFree(&grad);
    }
#endif
    
    phgDofFree(&f_h);
    phgDofFree(&u_t);
    phgDofFree(&u_h);
    phgDofFree(&ax);
    phgDofFree(&error);
    phgDofFree(&point);
    phgFreeGrid(&g);
    phgFinalize();
    return 0;
}
