/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2007 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: poisson.c,v 1.286 2007/09/27 07:37:20 zlb Exp $ */

#include "phg.h"

#include <string.h>
#include <math.h>

/*
    This test code solves the Poisson equation with Dirichlet BC:

	- \Delta u = f
	u |_{\partial\Omega} = g
*/

/* functions for analytical solution, BC and RHS */
#ifndef TEST_CASE
# define TEST_CASE 3
#endif
#include "functions.h"

/* constants for evaluating L2 error indicator */
#define C0L2	1.0
#define C1L2	0.5
/* constants for evaluating H1 error indicator */
#define C0H1	1.0
#define C1H1	0.5

#define FUNC_DEF(func_f, f_)				\
static void						\
func_f(FLOAT x, FLOAT y, FLOAT z, FLOAT *value)		\
{							\
    *value = f_(x, y, z);				\
    return;						\
}

FUNC_DEF(func_f, f_)
FUNC_DEF(func_g, g_)
FUNC_DEF(func_u, u_)

static void
func_grad(FLOAT x, FLOAT y, FLOAT z, FLOAT *value)
{
    grad_u_(value, x, y, z);
}

static double
elapsed_time(GRID *g, BOOLEAN flag, double mflops)
/* returns and prints elapsed time since last call to this function */
{
    static double t0 = 0.0;
    double et, tt[3];
    size_t mem;
 
    phgGetTime(tt);
    et = tt[2] - t0;
    t0 = tt[2];

    mem = phgMemoryUsage(g, NULL);

    if (flag) {
	if (mflops <= 0)
	    phgPrintf("[%0.4lgMB %0.4lfs]\n", mem / (1024.0 * 1024.0), et);
	else
	    phgPrintf("[%0.4lgMB %0.4lfs %0.4lgGF]\n", mem / (1024.0 * 1024.0),
			 et, mflops*1e-3);
    }

    return et;
}

#if 0
/* Note: the following code is only used to test the solver interface for
 *	 the case u_h->dim > 1 */
static void
build_linear_system(SOLVER *solver, DOF *u_h, DOF *f_h)
{
    int N = u_h->type->nbas;    /* number of basis functions in an element */
    int dim = u_h->dim;         /* DOF dimension */
    GRID *g = u_h->g;
    SIMPLEX *e;
    int i, j, k;
    FLOAT A[N][N], B[N][dim], buffer[N], *row;
    INT I[N][dim], J[dim][N];

    ForAllElements(g, e) {
	/* compute \int \grad\phi_i \cdot \grad\phi_j making use of symmetry */
	for (i = 0; i < N; i++) {
	    for (k = 0; k < dim; k++)
		I[i][k] = J[k][i] = phgSolverMapE2L(solver, 0, e, i * dim + k);
	    for (j = 0; j <= i; j++)
		A[j][i] = A[i][j] =
		    phgQuadGradBasDotGradBas(e, u_h, i, u_h, j, QUAD_DEFAULT);
	}

	/* loop on basis functions */
	for (i = 0; i < N; i++) {
	    if (phgDofDirichletBC(u_h, e, i, func_g,
				  buffer, B[i], DOF_PROJ_NONE)) {
		row = buffer;
	    }
	    else if (phgDofGetElementBoundaryType(u_h, e, i * dim) & NEUMANN) {
		/* TODO */
		row = NULL;
		phgError(1, "%s:%d: unimplemented.\n", __FILE__, __LINE__);
	    }
	    else {
		/* right hand side: \int f * phi_i */
		phgQuadDofTimesBas(e, f_h, u_h, i, QUAD_DEFAULT, B[i]);
		row = A[i];
	    }
	    for (k = 0; k < dim; k++)
		phgSolverAddMatrixEntries(solver, 1, &I[i][k], N, J[k], row); 
	}
	phgSolverAddRHSEntries(solver, N * dim, I[0], B[0]);
    }
}
#else
static void
build_linear_system(SOLVER *solver, DOF *u_h, DOF *f_h)
{
    int N = u_h->type->nbas;    /* number of basis functions in an element */
    GRID *g = u_h->g;
    SIMPLEX *e;
    int i, j;
    FLOAT A[N][N], B[N], buffer[N], *row;
    INT I[N];

    assert(u_h->dim == 1);

    ForAllElements(g, e) {
	/* compute \int \grad\phi_i \cdot \grad\phi_j making use of symmetry */
	for (i = 0; i < N; i++) {
	    I[i] = phgSolverMapE2L(solver, 0, e, i);
	    for (j = 0; j <= i; j++)
		A[j][i] = A[i][j] =
		    phgQuadGradBasDotGradBas(e, u_h, i, u_h, j, QUAD_DEFAULT);
	}

	/* loop on basis functions */
	for (i = 0; i < N; i++) {
	    if (phgDofDirichletBC(u_h, e, i, func_g,
				  buffer, B + i, DOF_PROJ_NONE)) {
		row = buffer;
	    }
	    else if (phgDofGetElementBoundaryType(u_h, e, i) & NEUMANN) {
		/* TODO */
		row = NULL;
		phgError(1, "%s:%d: unimplemented.\n", __FILE__, __LINE__);
	    }
	    else {
		/* right hand side: \int f * phi_i */
		phgQuadDofTimesBas(e, f_h, u_h, i, QUAD_DEFAULT, B + i);
		row = A[i];
	    }
	    phgSolverAddMatrixEntries(solver, 1, I + i, N, I, row); 
	}
	phgSolverAddRHSEntries(solver, N, I, B);
    }
}
#endif	/* 0 */

static void
estimate_error(DOF *u_h, DOF *f_h, DOF *grad_u, DOF *e_L2, DOF *e_H1)
/* compute L2 and H1 error indicators */
{
    int i;
    GRID *g = u_h->g;
    SIMPLEX *e;
    DOF *jump, *residual, *tmp;
    FLOAT eta0, eta1, d, h;
    FLOAT diam;

    jump = phgQuadFaceJump(grad_u, DOF_PROJ_DOT, "jumps", QUAD_DEFAULT);
    tmp = phgDofDivergence(grad_u, NULL, NULL, NULL);
    residual = phgDofCopy(f_h, NULL, DOF_DGn[u_h->type->order], NULL);
    phgDofAXPY(1.0, tmp, &residual);
    phgDofFree(&tmp);

    ForAllElements(g, e) {
	diam = phgGeomGetDiameter(g, e);
	e->mark = 0;		/* clear refinement mmark */
	eta0 = eta1 = 0.0;
	/* for each face F compute [grad_u \cdot n] */
	for (i = 0; i < NFace; i++) {
	    if (e->bound_type[i] & (DIRICHLET | NEUMANN))
		continue;	/* boundary face */
	    h = phgGeomGetFaceDiameter(g, e, i);
	    d = *DofFaceData(jump, e->faces[i]);
	    eta0 += d * h * h * h;
	    eta1 += d * h;
	}
	d = phgQuadDofDotDof(e, residual, residual, QUAD_DEFAULT);
	eta0 = eta0 * C0L2 + diam * diam * diam * diam * d * C1L2;
	eta1 = eta1 * C0H1 + diam * diam * d * C1H1;

	/* add curved boundary errors (FIXME: how to normalize?) */
	d = phgGetBoundaryError(g, e);
	eta0 += d;
	eta1 += d;

	*DofElementData(e_L2, e->index) = Sqrt(eta0);
	*DofElementData(e_H1, e->index) = Sqrt(eta1);
    }
    phgDofFree(&jump);
    phgDofFree(&residual);

    return;
}

static void
mark_refine(DOF *indicator, FLOAT thres, int depth)
{
    GRID *g = indicator->g;
    SIMPLEX *e;

    ForAllElements(g, e)
	e->mark = (*DofElementData(indicator, e->index) >= thres) ? depth : 0;
}

int
main(int argc, char *argv[])
{
    GRID *g;
    DOF *u_h, *f_h, *grad_u = NULL, *e_L2, *e_H1;
    SOLVER *solver;
    FLOAT L2_error = 0.0, H1_error = 0.0, EL2_error = 0.0, EH1_error = 0.0;
    INT nits;
    BOOLEAN flag;
    size_t mem, mem_peak;

    /*------------------- Command-line options ----------------------*/

    /* refine strategies */
    static const char *strategies[] = {"maximum", "equidist", NULL};
    typedef enum {MAXIMUM, EQUIDIST} STRATEGY;
    static int strategy = MAXIMUM;

    /* List of valid export functions */
    static const char *export_names[] = {"vtk", "opendx", NULL};
    static const char *(*export_list[])(GRID *, const char *, DOF *, ...) =
				{phgExportVTK, phgExportDX};
    static int export_index = 0;
    /* file extensions */
    static const char *export_ext[] = {"vtk", "dx"};
    static FLOAT lif_threshold = 1.2;
    static INT submesh_threshold = 1024;

    /* Scalar parameters */
    static BOOLEAN	refine_depth = 1;
    static FLOAT	tol = 4e-3;
    static INT		mem_max = 400;
    static BOOLEAN	export_mesh = FALSE;
    static BOOLEAN	export_all = FALSE;
    static BOOLEAN	dump_matrix = FALSE;
    static BOOLEAN	dump_rhs = FALSE;
    static BOOLEAN	L2adapt = FALSE;
    static BOOLEAN	use_weights = FALSE;

    phgOptionsRegisterFilename("mesh_file", "Mesh file", (char **)&fn);
    phgOptionsRegisterInt("pre_refines", "Pre-refines", &pre_refines);
    phgOptionsRegisterInt("refine_depth",
			  "Minimum refine depth (<=0 ==> uniform refinement)",
			  &refine_depth);
    phgOptionsRegisterKeyword("refine_strategy", "Adative strategy",
			 strategies, &strategy);
    phgOptionsRegisterFloat("tol", "Convergence criterion", &tol);
    phgOptionsRegisterFloat("lif_threshold", "Load imbalance factor threshold",
				&lif_threshold);
    phgOptionsRegisterInt("submesh_threshold", "Submesh size threshold",
				&submesh_threshold);
    phgOptionsRegisterInt("mem_max", "Maximum memory (MB)", &mem_max);
    phgOptionsRegisterNoArg("export_mesh", "Export mesh", &export_mesh);
    phgOptionsRegisterNoArg("export_all", "Export all meshes", &export_all);
    phgOptionsRegisterKeyword("export_format", "Format of Export files",
			export_names, &export_index);
    phgOptionsRegisterNoArg("dump_matrix", "Dump matrix", &dump_matrix);
    phgOptionsRegisterNoArg("dump_rhs", "Dump RHS", &dump_rhs);
    phgOptionsRegisterNoArg("L2adapt", "adaptation using L2 error indicator",
				&L2adapt);
    phgOptionsRegisterNoArg("use_weights", "use error indicator as weights",
				&use_weights);

    /*------------------- End command-line options ----------------------*/

    phgOptionsPreset("-default_dof_type P3");
    phgVerbosity = 0;
    phgInit(&argc, &argv);

    g = phgNewGrid(-1);
    if (!phgImport(g, fn, FALSE))
	phgError(1, "can't read file \"%s\".\n", fn);

    /* pre-refine */
    phgRefineAllElements(g, pre_refines);

    u_h = phgDofNew(g, DOF_DEFAULT, 1, "u_h", DofInterpolation);
    phgDofSetDataByValue(u_h, 1.0);
#if 0
    f_h = phgDofNew(g, DOF_ANALYTIC, 1, "f_h",  func_f);
#else
    f_h = phgDofNew(g, DOF_DEFAULT, 1, "f_h",  func_f);
#endif
    e_L2 = phgDofNew(g, DOF_P0, 1, "L2 error indicator", DofNoAction);
    e_H1 = phgDofNew(g, DOF_P0, 1, "H1 error indicator", DofNoAction);

    phgPrintf("Test case %d, mesh file: \"%s\"\n", TEST_CASE, fn);
    phgPrintf("Analytic solution: \"%s\"\n", analytic);
    phgPrintf("Approximation: u_h = %s, f_h = %s\n\n", u_h->type->name,
		f_h->type->name);
    phgOptionsShowUsed();		/* show options */

    while (TRUE) {
	elapsed_time(g, FALSE, 0.);	/* reset timer */
	phgPrintf("\n*** Pass %d\n", g->serial_no % 65536);
#if DEBUG
	phgPrintf("--- Check grid: ");
	phgCheckConformity(g);
	elapsed_time(g, TRUE, 0.);
#endif
	phgPrintf("DOF: %d, T: %d, V: %d, E: %d, F: %d\n",
		  DofGetDataCountGlobal(u_h),
		  g->nleaf_global, g->nvert_global, g->nedge_global,
		  g->nface_global);
	if (!use_weights && phgBalanceGrid(g, lif_threshold, submesh_threshold,
						 NULL, 0.)) {
	    phgPrintf("Redistribute grid ");
	    elapsed_time(g, TRUE, 0.);
#if DEBUG
	    phgPrintf("--- Check grid: ");
	    phgCheckConformity(g);
	    elapsed_time(g, TRUE, 0.);
#endif
	}
	phgPrintf("Grid distribution: %d submesh%s, LIF = %0.3lg\n",
			g->nprocs, g->nprocs > 1 ? "es" : "", (double)g->lif);

	phgPrintf("Prepare solver: ");
	if ((solver = phgSolverCreate(SOLVER_DEFAULT, u_h, NULL)) == NULL) {
	    phgPrintf("abort.\n");
	    break;
	}
	phgPrintf("LIF %lg ", (double)solver->mat->rmap->lif);
	elapsed_time(g, TRUE, 0.);
    
	phgPrintf("Build linear system: ");
	phgPerfGetMflops(g, NULL, NULL);	/* reset flops counter */
	build_linear_system(solver, u_h, f_h);
	elapsed_time(g, TRUE, phgPerfGetMflops(g, NULL, NULL));

#if 0
/* test block matrix */
{
    MAT *a = phgMatCreateBlockMatrix(g, 1, 1, &solver->mat, NULL);
    phgMatDestroy(&solver->mat);
    solver->mat = a;
}
#endif

#if 0
/* test solving multiple RHS with a same matrix */
{
    INT nlocal = solver->rhs->map->nlocal;
    INT osize = solver->rhs->map->localsize - nlocal;
    FLOAT *rhs = phgAlloc(nlocal * sizeof(*rhs));
    FLOAT *rhs_offp = phgAlloc(osize * sizeof(*rhs_offp));
    FLOAT *data = phgAlloc(DofGetDataCount(u_h) * sizeof(*u_h->data));
    /* save (unassembled) RHS and initial solution */
    memcpy(rhs, solver->rhs->data, nlocal * sizeof(*rhs));
    if (osize > 0)
	memcpy(rhs_offp, solver->rhs->offp_data, osize * sizeof(*rhs_offp));
    memcpy(data, u_h->data, DofGetDataCount(u_h) * sizeof(*u_h->data));
    /*bzero(solver->rhs, localsize * sizeof(*rhs));*/
    phgSolverSolve(solver, FALSE, u_h, NULL);
    /* reset RHS */
    phgSolverResetRHS(solver);
    memcpy(solver->rhs->data, rhs, nlocal * sizeof(*rhs));
    if (osize > 0)
	memcpy(solver->rhs->offp_data, rhs_offp, osize * sizeof(*rhs_offp));
    memcpy(u_h->data, data, DofGetDataCount(u_h) * sizeof(*u_h->data));
    phgFree(rhs);
    phgFree(rhs_offp);
    phgFree(data);
}
#endif

	phgPrintf("Assemble linear system: ");
	phgSolverAssemble(solver);
	elapsed_time(g, TRUE, 0.);

	if (dump_matrix || dump_rhs) {
	    static char fn0[64], fn1[64];
	    static int no = 0;
	    sprintf(fn0, "matrix%d.dat", no);
	    sprintf(fn1, "rhs%d.dat", no++);
	    phgMatDumpCSR(solver->mat, dump_matrix ? fn0 : NULL);
	    phgVecDump(solver->rhs, dump_rhs ? fn1 : NULL);
	}

	phgPrintf("Solve linear system: ");
	phgPerfGetMflops(g, NULL, NULL);	/* reset flops counter */
	if ((nits = phgSolverSolve(solver, TRUE, u_h, NULL)) < 0) {
	    phgPrintf("abort.\n");
	    break;
	}
	phgPrintf("%d its, %0.4le ", nits, (double)solver->residual);
	elapsed_time(g, TRUE, phgPerfGetMflops(g, NULL, NULL));

	phgSolverDestroy(&solver);

	phgPrintf("Estimate: ");
	grad_u = phgDofGradient(u_h, NULL, NULL, "grad u_h");
	estimate_error(u_h, f_h, grad_u, e_L2, e_H1);
	EL2_error = phgDofNormL2Vec(e_L2);
	EH1_error = phgDofNormL2Vec(e_H1);
	phgPrintf("EL2=%0.2le EH1=%0.2le ", (double)EL2_error,
					    (double)EH1_error);
	mem = phgMemoryUsage(g, &mem_peak);
	flag = (EH1_error < tol || mem_peak > 1024 * (size_t)mem_max * 1024);

	if (has_analytic_solution) {
	    DOF *u, *grad;
	    u = phgDofNew(g, u_h->type, u_h->dim, "analytic solution", func_u);
	    grad = phgDofNew(g, u_h->type->grad_type, Dim * u_h->dim,
				"analytic gradient", func_grad);
	    phgDofAXPY(-1.0, u_h, &u);
	    /* Note: the following works only if u is smooth! */
	    phgDofAXPY(-1.0, grad_u, &grad);
	    L2_error = phgDofNormL2(u);
	    H1_error = phgDofNormL2(grad);
	    phgPrintf("L2=%0.2le H1=%0.2le ", (double)L2_error,
			    		      (double)H1_error);
	    phgDofFree(&grad);
	    phgDofFree(&u);
	}
	elapsed_time(g, TRUE, 0.);

	phgPrintf("Memory usage: current %0.4lgMB, peak %0.4lgMB\n",
		(double)mem / (1024.0 * 1024.0),
		(double)mem_peak / (1024.0 * 1024.0));

    	if (export_all || (flag && export_mesh)) {
	    static char fn[64];
	    const char *p;
	    if (export_all)
		sprintf(fn, "mesh%03d.%s", g->serial_no % 65536,
				export_ext[export_index]);
	    else
		sprintf(fn, "mesh.%s", export_ext[export_index]);
	    phgPrintf("Export %s file %s ",
			    export_names[export_index],
			    p = export_list[export_index](g, fn, u_h, grad_u,
							e_L2, e_H1, NULL));
            if (phgRank == 0) {
                FILE *f;
                long long size;
                f = fopen(p, "r");
                fseek(f, 0L, SEEK_END);
                size = ftell(f);
                fclose(f);
                phgPrintf("size = %lld ", size);
            }
	    elapsed_time(g, TRUE, 0.);
	    /*phgExportALBERT(g, "mesh.dat");*/
    	}

	phgDofFree(&grad_u);

#define Idle(g)	((g) == NULL || phgRank >= (g)->nprocs)
	if (flag && !Idle(g))
	    break;

	if (use_weights && phgBalanceGrid(g, lif_threshold, submesh_threshold,
				L2adapt ? e_L2 : e_H1, 1.)) {
	    phgPrintf("Redistribute grid, %d submeshes, LIF = %lg ",
			g->nprocs, (double)g->lif);
	    elapsed_time(g, TRUE, 0.);
#if DEBUG
	    phgPrintf("--- Check grid: ");
	    phgCheckConformity(g);
	    elapsed_time(g, TRUE, 0.);
#endif
	}

	phgPrintf("Refine mesh: ");
	if (refine_depth <= 0) {
	    /* uniform refine */
	    phgRefineAllElements(g, refine_depth == 0 ? 3 : -refine_depth);
	}
	else {
	    DOF *ei = L2adapt ? e_L2 : e_H1;
	    /* adaptive refine */
	    switch ((STRATEGY)strategy) {
		case MAXIMUM:
		    /* maximum strategy */
		    mark_refine(ei, 0.5 * phgDofNormInftyVec(ei), refine_depth);
		    break;
		case EQUIDIST:
		    /* equidistribution strategy */
		    mark_refine(ei, tol / Sqrt((double)g->nleaf_global),
				refine_depth);
		    break;
	    }
	    phgRefineMarkedElements(g);
	}
	elapsed_time(g, TRUE, 0.);

	if (has_analytic_solution) {
	    DOF *u0 = phgDofNew(g, u_h->type, u_h->dim, "u0", func_u);
	    phgDofAXPY(-1.0, u_h, &u0);
	    phgPrintf("Interpolation error: %0.3le\n",
			    (double)phgDofNormInftyVec(u0));
	    phgDofFree(&u0);
	}
    }

    phgDofFree(&u_h);
    phgDofFree(&f_h);
    phgDofFree(&e_L2);
    phgDofFree(&e_H1);
    phgFreeGrid(&g);

    phgFinalize();

    return 0;
}
