/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2007 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* AMS, Auxiliary space Maxwell Solver, solves the following time-harmonic
 * Maxwell equation:
 *
 *	curl(alpha curl) E + beta E = J
 *
 * with alpha > 0 and beta >= 0, discretized using edge elements.
 *
 * $Id: solver-ams.c,v 1.50 2007/09/27 07:37:20 zlb Exp $
 */

#define TEST_HIGH_ORDER	0	/* 1 ==> test code for higher orders */

#define HYPRE_TEST	0	/* 1 ==> test with Hypre AMS preconditioner */

#include "phg.h"

#include <string.h>
#include <strings.h>
#include <math.h>
#include <stdlib.h>

typedef struct {
    DOF		*alpha, *beta;	/* The coefficients alpha and beta */
    MAT		*G, *Pi[3];	/* The G and Pi matrices (P1) */
    MAT		*A_G, *A_Pi;	/* The Poisson matrices (P1) */
    SOLVER	*solver_G, *solver_Pi;

#if TEST_HIGH_ORDER
    MAT		*G0, *Pi0[3];	/* The G and Pi matrices (original order) */
    MAT		*A_G0, *A_Pi0;	/* The Poisson matrices (original order) */
#endif	/* TEST_HIGH_ORDER */

#if HYPRE_TEST
    SOLVER	*solver_hypre;
#endif	/* HYPRE_TEST */
    BOOLEAN	assembled;
    BOOLEAN	coefficients_defined;	/* this flag is not really needed */

    char	*cycle_type;

    int		aux_solver_index;
    char	*aux_solver_opts;
    FLOAT	aux_solver_rtol;
    INT		aux_solver_maxit;
} OEM_DATA;

static char *cycle_type = "02120";	/* default cycle type */

static int	aux_solver_index	= -1;
static char	*aux_solver_opts	= NULL;
static FLOAT	aux_solver_rtol		= 0.0;
static INT	aux_solver_maxit	= 1;

static int	grad_solver_index	= 0;
static char	*grad_solver_opts	= "-pcg_pc_type jacobi";
static FLOAT	grad_solver_rtol	= 1e-16;
static INT	grad_solver_maxit	= 2000;

/* convenience macro */
#define _t	((OEM_DATA *)solver->oem_data)

#define Initialize		NULL
#define Finalize		NULL
#define AddMatrixEntries	NULL
#define AddRHSEntries		NULL

/*---------------------------- auxiliary functions -------------------------*/

static void
gradient_build(SOLVER *solver, DOF *u_h, DOF *f_h, DOF *beta)
{
    int N = u_h->type->nbas;    /* number of basis functions in an element */
    GRID *g = u_h->g;
    SIMPLEX *e;
    int i, j;
    FLOAT A[N][N], B[N], buffer[N], *row;
    INT I[N];

    assert(u_h->dim == 1);

    ForAllElements(g, e) {
	/* Note: only works with piecewise constant beta */
	if (beta != NULL && *DofElementData(beta, e->index) != 0.0) {
	    for (i = 0; i < N; i++) {
		j = phgSolverMapE2L(solver, 0, e, i);
		phgSolverAddMatrixEntry(solver, j, j, 1.0);
		phgSolverAddRHSEntry(solver, j, 0.0);
	    }
	    continue;
	}

	/* compute \int \grad\phi_i \cdot \grad\phi_j making use of symmetry */
	for (i = 0; i < N; i++) {
	    I[i] = phgSolverMapE2L(solver, 0, e, i);
	    for (j = 0; j <= i; j++)
		A[j][i] = A[i][j] =
		    phgQuadGradBasDotGradBas(e, u_h, i, u_h, j, QUAD_DEFAULT);
	}

	/* loop on basis functions */
	for (i = 0; i < N; i++) {
	    if (phgDofDirichletBC(u_h, e, i, NULL,
				  buffer, B+ i, DOF_PROJ_NONE)) {
		row = buffer;
	    }
	    else if (phgDofGetElementBoundaryType(u_h, e, i) & NEUMANN) {
		/* TODO */
		row = NULL;
		phgError(1, "%s:%d: unimplemented.\n", __FILE__, __LINE__);
	    }
	    else {
		/* right hand side: \int f * phi_i */
		B[i] = phgQuadDofDotGradBas(e, f_h, u_h, i, QUAD_DEFAULT);
		row = A[i];
	    }
	    phgSolverAddMatrixEntries(solver, 1, I + i, N, I, row); 
	}
	phgSolverAddRHSEntries(solver, N, I, B);
    }
}

void
phgSolverAMSRemoveGradient(DOF **fptr, DOF *beta, int order)
/* Returns a new DOF which is equal to f - \grad\psi, where \psi is 0 in
 * the region beta != 0, and satisfies the following poisson equation:
 * 	<\grad\psi, \grad\phi> = <f, \grad\phi>
 * in the region beta == 0.
 *
 * beta == NULL means beta == 0 everywhere */
{
    SOLVER *solver;
    DOF *f, *psi, *grad;
    INT verbosity = phgVerbosity;

    assert(fptr != NULL && *fptr != NULL);

    assert(beta == NULL || beta->type == DOF_P0 || beta->type == DOF_DG0
	   || beta->type == DOF_CONSTANT);

    f = *fptr;

    assert(f->type != DOF_ANALYTIC && f->type != DOF_CONSTANT
		&& DofDim(f) == Dim && order >= 1 && order <= 4);

    psi = phgDofNew(f->g, DOF_Pn[order], 1, "psi", DofNoAction);

    /* Set up the Poisson solver */
    phgOptionsPush();
    phgOptionsSetOptions(grad_solver_opts);
    phgOptionsSetFloat("-solver_rtol", grad_solver_rtol);
    phgOptionsSetInt("-solver_maxit", grad_solver_maxit);
    solver = phgSolverCreate(phgSolverList[grad_solver_index], psi, NULL);
    phgOptionsPop();

    gradient_build(solver, psi, f, beta);
    phgVerbosity--;
    phgSolverSolve(solver, TRUE, psi, NULL);
    phgVerbosity = verbosity;
    if (phgVerbosity > 0)
	phgPrintf("Solve gradient equation: nits = %d, residual = %0.2le ",
			solver->nits, (double)solver->residual);
    phgSolverDestroy(&solver);

    grad = phgDofGradient(psi, NULL, NULL, NULL);
    phgDofFree(&psi);
    phgDofAXPY(-1.0, grad, fptr);
    phgDofFree(&grad);

    return;
}

void
phgSolverAMSSetCoefficients(SOLVER *solver, DOF *alpha, DOF *beta)
/* defines the coefficients alpha and beta (NULL means 1) */
{
    if (_t->alpha != NULL)
	phgDofFree(&_t->alpha);
    if (_t->beta != NULL)
	phgDofFree(&_t->beta);

    if (alpha != NULL)
	_t->alpha = phgDofCopy(alpha, NULL, NULL, NULL);
    if (beta != NULL)
	_t->beta = phgDofCopy(beta, NULL, NULL, NULL);

    _t->coefficients_defined = TRUE;
}

void
phgSolverAMSSetConstantCoefficients(SOLVER *solver, FLOAT a, FLOAT b)
{
    GRID *g = solver->mat->rmap->dofs[0]->g;

    if (_t->alpha != NULL)
	phgDofFree(&_t->alpha);
    if (_t->beta != NULL)
	phgDofFree(&_t->beta);

    if (a != 1.0) {
	_t->alpha = phgDofNew(g, DOF_CONSTANT, 1, "alpha", DofNoAction);
	phgDofSetDataByValue(_t->alpha, a);
    }

    if (b != 1.0) {
	_t->beta = phgDofNew(g, DOF_CONSTANT, 1, "beta", DofNoAction);
	phgDofSetDataByValue(_t->beta, b);
    }

    _t->coefficients_defined = TRUE;
}

static DOF *dof_H1;

static void
func_G(DOF *u, SIMPLEX *e, int bno, const FLOAT *lambda, FLOAT *values)
/* dummy function for computing G.
 * Note: the results are stored as values[u->dim][Dim] */
{
    int i;
    const FLOAT *p, (*J)[Dim + 1];

    Unused(bno);
    /* TODO: cache grad_lambda of the bases of dof_H1 */
    p = dof_H1->type->BasGrads(dof_H1, e, 0, -1, lambda);
    J = (void *)phgGeomGetJacobian(u->g, e);
    for (i = 0; i < u->dim; i++, values += Dim, p += Dim + 1) {
	values[0] = p[0]*J[0][0] + p[1]*J[1][0] + p[2]*J[2][0] + p[3]*J[3][0];
	values[1] = p[0]*J[0][1] + p[1]*J[1][1] + p[2]*J[2][1] + p[3]*J[3][1];
	values[2] = p[0]*J[0][2] + p[1]*J[1][2] + p[2]*J[2][2] + p[3]*J[3][2];
    }
}

static void
func_Pi(DOF *u, SIMPLEX *e, int bno, const FLOAT *lambda, FLOAT *values)
/* dummy function for computing Pi.
 * Note: the results are stored as values[u->dim][Dim] */
{
    int i;
    const FLOAT *p;

    Unused(bno);
    /* TODO: cache bases of dof_H1 */
    p = dof_H1->type->BasFuncs(dof_H1, e, 0, -1, lambda);
    for (i = 0; i < u->dim / Dim; i++, values += 3 * Dim, p++) {
	values[0] = *p;
	values[1] = 0.;
	values[2] = 0.;

	values[3] = 0.;
	values[4] = *p;
	values[5] = 0.;

	values[6] = 0.;
	values[7] = 0.;
	values[8] = *p;
    }
}

static void
build_G_Pi(SOLVER *solver, int order, MAT **G, MAT *Pi[])
/* builds the gradient matrix G whose elements are determined by:
 *	\grad\phi_j = \sum g_{kj} \psi_k
 * where \phi_j is the j-th basis function of the H1 space,
 * and \psi_k is the k-th basis function of the Hcurl space */
{
    MAP *map = solver->mat->rmap, *map_H1;
    DOF *u = map->dofs[0];
    GRID *g = u->g;
    SIMPLEX *e;
    DOF *dof_Hcurl;
    DOF *dof_Hcurl3;
    DOF_TYPE *h1_type = NULL;
    INT i, j, N_Hcurl, N_H1, *cols, *rows;
    FLOAT *p, *buffer, *buffer0, *buffer1, *buffer2, *pdata[15];

    assert(map->ndof == 1 && u->type->dim == 3);

    h1_type = (order <= 4 ? DOF_Pn : DOF_HBn)[order];

    N_H1 = h1_type->nbas;
    N_Hcurl = u->type->nbas;

    dof_H1 = phgDofNew(g, h1_type, 1, "tmp", DofNoData);
    map_H1 = phgMapCreate(dof_H1, NULL);

    dof_Hcurl = phgDofNew(g, u->type, N_H1, "tmp", DofNoData);
    *G = phgMapCreateMat(map, map_H1);
    (*G)->assembled = FALSE;

    dof_Hcurl3 = phgDofNew(g, u->type, Dim * N_H1, "tmp3", DofNoData);
    Pi[0] = phgMapCreateMat(map, map_H1);
    Pi[0]->assembled = FALSE;
    Pi[1] = phgMapCreateMat(map, map_H1);
    Pi[1]->assembled = FALSE;
    Pi[2] = phgMapCreateMat(map, map_H1);
    Pi[2]->assembled = FALSE;

    buffer = phgAlloc(2 * Dim * N_Hcurl * N_H1 * sizeof(*buffer));
    buffer0 = buffer  + Dim * N_Hcurl * N_H1;
    buffer1 = buffer0 + N_Hcurl * N_H1;
    buffer2 = buffer1 + N_Hcurl * N_H1;

    rows = phgAlloc((N_Hcurl + N_H1) * sizeof(*rows));
    cols = rows  + N_Hcurl;

    ForAllElements(g, e) {
	for (j = 0; j < N_Hcurl; j++)
	    rows[j] = phgMapE2L(map, 0, e, j);
	for (j = 0; j < N_H1; j++)
	    cols[j] = phgMapE2L(map_H1, 0, e, j);

	/* The gradient matrix */
	p = buffer;
	j = dof_Hcurl->dim * dof_Hcurl->type->np_edge;
	for (i = 0; i < NEdge; i++) {
	    u->type->InitFunc(dof_Hcurl, e, EDGE, i, NULL, func_G,
				NULL, p, pdata);
	    pdata[4 + i] = p;
	    p += j;
	}

	j = dof_Hcurl->dim * dof_Hcurl->type->np_face;
	if (j > 0) {
	    for (i = 0; i < NFace; i++) {
		u->type->InitFunc(dof_Hcurl, e, FACE, i, NULL, func_G,
					NULL, p, pdata);
		pdata[10 + i] = p;
		p += j;
	    }
	}

	if (dof_Hcurl->type->np_elem > 0)
	    u->type->InitFunc(dof_Hcurl, e, ELEMENT, 0, NULL, func_G,
				NULL, p, pdata);

	phgMatSetEntries(*G, N_Hcurl, rows, N_H1, cols, buffer);

	/* The (H1)^3 matrix */
	p = buffer;
	j = dof_Hcurl3->dim * dof_Hcurl3->type->np_edge;
	for (i = 0; i < NEdge; i++) {
	    u->type->InitFunc(dof_Hcurl3, e, EDGE, i, NULL, func_Pi,
				NULL, p, pdata);
	    pdata[4 + i] = p;
	    p += j;
	}

	j = dof_Hcurl3->dim * dof_Hcurl3->type->np_face;
	if (j > 0) {
	    for (i = 0; i < NFace; i++) {
		u->type->InitFunc(dof_Hcurl3, e, FACE, i, NULL, func_Pi,
					NULL, p, pdata);
		pdata[10 + i] = p;
		p += j;
	    }
	}

	if (dof_Hcurl3->type->np_elem > 0)
	    u->type->InitFunc(dof_Hcurl3, e, ELEMENT, 0, NULL, func_Pi,
				NULL, p, pdata);

	p = buffer;
	for (i = 0; i < N_Hcurl; i++) {
	    for (j = 0; j < N_H1; j++) {
		buffer0[i * N_H1 + j] = *(p++);
		buffer1[i * N_H1 + j] = *(p++);
		buffer2[i * N_H1 + j] = *(p++);
	    }
	}

	phgMatSetEntries(Pi[0], N_Hcurl, rows, N_H1, cols, buffer0);
	phgMatSetEntries(Pi[1], N_Hcurl, rows, N_H1, cols, buffer1);
	phgMatSetEntries(Pi[2], N_Hcurl, rows, N_H1, cols, buffer2);
    }	/* ForAllElements */

    phgMatAssemble(*G);
    phgMatAssemble(Pi[0]);
    phgMatAssemble(Pi[1]);
    phgMatAssemble(Pi[2]);

    phgFree(rows);
    phgFree(buffer);

    phgMapDestroy(&map_H1);
    phgDofFree(&dof_H1);
    phgDofFree(&dof_Hcurl);
    phgDofFree(&dof_Hcurl3);
    
    return;
}

static void
build_poisson(SOLVER *solver, int order, DOF *alpha, DOF *beta,
					 MAT **ma, MAT **mb)
/* builds the Poisson matrices:
 *	ma <==> div(alpha grad) + beta
 *	mb <==> div(beta grad)
 * alpha==NULL means alpha=1, beta==NULL means beta=1 */
{
    DOF *tmp, *u;
    DOF_TYPE *h1_type;
    MAP *map;
    int i, j, n, N;
    GRID *g;
    SIMPLEX *e;
    FLOAT *A, *B = NULL, *A1, *B1 = NULL, *A0, d;
    INT *I, *J;
    BTYPE *b, DB_mask;

    u = solver->mat->rmap->dofs[0];
    g = u->g;

    /* TODO: check DOF_TYPE of solver->dofs[] for consistency */
    assert(solver->mat->rmap->ndof == 1 && DofTypeDim(u->type) == 3 &&
	   u->dim == 1);
    assert(beta == NULL || beta->dim == 1);
    assert(solver->mat->rmap->dofs != NULL);

    DB_mask = solver->mat->rmap->dofs[0]->DB_mask;

    h1_type = (order <= 4 ? DOF_Pn : DOF_HBn)[order];
    tmp = phgDofNew(g, h1_type, 1, "H1", DofNoData);
    map = phgMapCreate(tmp, NULL);
    *ma = phgMapCreateMat(map, map);
    (*ma)->assembled = FALSE;
    if (beta != NULL && beta->type == DOF_CONSTANT && *beta->data == 0.) {
	*mb = NULL;
    }
    else {
	*mb = phgMapCreateMat(map, map);
	(*mb)->assembled = FALSE;
    }
    phgMapDestroy(&map);

    N = tmp->type->nbas;	/* number of basis functions in an element */
    A = phgAlloc(N * N * sizeof(*A));
    A0 = phgAlloc(N * N * sizeof(*A0));
    A1 = phgAlloc(N * sizeof(*A1));
    if (*mb != NULL) {
	B = phgAlloc(N * N * sizeof(*B));
	B1 = phgAlloc(N * sizeof(*B1));
    }
    I = phgAlloc(N * sizeof(*I));
    J = phgAlloc(N * sizeof(*J));
    b = phgAlloc(N * sizeof(*b));

    /* Note: if beta is constant then the element mass matrices only vary
       by element volumes */
    if (beta == NULL || beta->type == DOF_CONSTANT) {
	e = g->roots;
	d = (beta == NULL ? 1.0 : *beta->data) / phgGeomGetVolume(g, e);
	for (i = 0; i < N; i++)
	    for (j = 0; j <= i; j++)
		A0[i * N + j] = A0[j * N + i] =
		    d * phgQuadBasDotBas(e, tmp, i, tmp, j, -1);
    }

    ForAllElements(g, e) {
	if (beta != NULL && beta->type != DOF_CONSTANT) {
	    /* compute \int beta * phi_i * phi_j */
	    for (i = 0; i < N; i++)
		for (j = 0; j <= i; j++)
		    A[i * N + j] = phgQuadBasABas(e, tmp, i, beta, tmp, j, -1);
	}
	else {
	    d = phgGeomGetVolume(g, e);
	    for (i = 0; i < N; i++)
		for (j = 0; j <= i; j++)
		    A[i * N + j] = A0[i * N + j] * d;
	}
	for (i = 0; i < N; i++) {
	    I[i] = phgMapE2L((*ma)->rmap, 0, e, i);
	    b[i] = phgDofGetElementBoundaryType(tmp, e, i * tmp->dim);
	    for (j = 0; j <= i; j++) {
		/* FIXME: use phgQuadGradBasDotGradBas if both alpha and beta
		 * are constant */
		A[j * N + i] = (A[i * N + j] +=
			phgQuadGradBasAGradBas(e, tmp, i, alpha, tmp, j, -1));
		if (*mb != NULL)
		    B[j * N + i] = B[i * N + j] =
			phgQuadGradBasAGradBas(e, tmp, i, beta, tmp, j, -1);
	    }
	}
	/* Stiffness matrix */
	for (i = 0; i < N; i++) {
	    if (b[i] & DB_mask) {
#if 0
		phgMatAddEntry(*ma, I[i], I[i], 1.0);
		if (*mb != NULL)
		    phgMatAddEntry(*mb, I[i], I[i], 1.0);
#else
		phgMatAddEntry(*ma, I[i], I[i], A[i * N + i]);
		if (*mb != NULL)
		    phgMatAddEntry(*mb, I[i], I[i], B[i * N + i]);
#endif
	    }
	    else {
		for (j = 0, n = 0; j < N; j++) {
		    if (b[j] & DB_mask)
			continue;
		    J[n] = I[j];
		    A1[n] = A[i * N + j];
		    if (*mb != NULL)
			B1[n] = B[i * N + j];
		    n++;
		}
		phgMatAddEntries(*ma, 1, I + i, n, J, A1);
		if (*mb != NULL)
		    phgMatAddEntries(*mb, 1, I + i, n, J, B1);
	    }
	}
    }	/* end-ForAllElement */

    phgMatAssemble(*ma);
    if (*mb != NULL)
	phgMatAssemble(*mb);

    phgFree(A);
    phgFree(A0);
    phgFree(B);
    phgFree(A1);
    phgFree(B1);
    phgFree(I);
    phgFree(J);
    phgFree(b);

    phgDofFree(&tmp);
}

/*----------------------------------------------------------------------*/

static int
RegisterOptions(void)
{
    phgOptionsRegisterTitle("\nBuilt-in AMS solver options:", NULL, "ams");

    phgOptionsRegisterString("ams_cycle_type", "AMS cycle type", &cycle_type);

    /* options for the aux solver */

#if USE_HYPRE && HYPRE_VERSION_MAJOR >= 2
    /* set default aux solver to Hypre BoomerAMG */
    for (aux_solver_index = 0;
	 phgSolverNames[aux_solver_index] != NULL
	 && strcmp(phgSolverNames[aux_solver_index], "hypre") != 0;
	 aux_solver_index++);
    assert(phgSolverNames[aux_solver_index] != NULL);
#endif	/* USE_HYPRE && HYPRE_VERSION_MAJOR >= 2 */

#if USE_TRILINOS
    if (aux_solver_index < 0 || phgSolverNames[aux_solver_index] == NULL) {
	for (aux_solver_index = 0;
	     phgSolverNames[aux_solver_index] != NULL
	     && strcmp(phgSolverNames[aux_solver_index], "hypre") != 0;
	     aux_solver_index++);
	assert(phgSolverNames[aux_solver_index] != NULL);
    }
#endif	/* USE_TRILINOS */

    /* Fall back to first available solver */
    if (aux_solver_index < 0)
	aux_solver_index = 0;

    if (phgSolverNames[aux_solver_index] == NULL)
	aux_solver_index = 0;

    phgOptionsRegisterKeyword("ams_aux_solver", "Auxiliary solver",
			      phgSolverNames, &aux_solver_index);
    phgOptionsRegisterString("ams_aux_solver_opts", "Auxiliary solver options",
			      &aux_solver_opts);
    phgOptionsRegisterFloat("ams_aux_solver_rtol", "Auxiliary solver rtol",
			      &aux_solver_rtol);
    phgOptionsRegisterInt("ams_aux_solver_maxit", "Auxiliary solver maxit",
			  &aux_solver_maxit);

    /* options for the Gradient solver */
    phgOptionsRegisterKeyword("ams_grad_solver", "Gradient solver",
			      phgSolverNames, &grad_solver_index);
    phgOptionsRegisterString("ams_grad_solver_opts", "Gradient solver options",
			      &grad_solver_opts);
    phgOptionsRegisterFloat("ams_grad_solver_rtol", "Gradient solver rtol",
			      &grad_solver_rtol);
    phgOptionsRegisterInt("ams_grad_solver_maxit", "Gradient solver maxit",
			  &grad_solver_maxit);

    return 0;
}

static int
Init(SOLVER *solver)
{
    solver->oem_data = phgCalloc(1, sizeof(OEM_DATA));

    /* copy relavant cmdline arguments */
    _t->cycle_type = strdup(cycle_type);

    _t->aux_solver_index = aux_solver_index;
    if (aux_solver_opts != NULL)
	_t->aux_solver_opts = strdup(aux_solver_opts);
    _t->aux_solver_rtol = aux_solver_rtol;
    _t->aux_solver_maxit = aux_solver_maxit;

    return 0;
}

static int
Create(SOLVER *solver)
{
    assert(solver->mat->type != PHG_MATRIX_FREE);
    return 0;
}

static int
Destroy(SOLVER *solver)
{
    if (solver->oem_data == NULL)
	return 0;

    if (_t->alpha != NULL)
	phgDofFree(&_t->alpha);

    if (_t->beta != NULL)
	phgDofFree(&_t->beta);

    if (_t->G != NULL)
	phgMatDestroy(&_t->G);
    if (_t->Pi[0] != NULL)
	phgMatDestroy(&_t->Pi[0]);
    if (_t->Pi[1] != NULL)
	phgMatDestroy(&_t->Pi[1]);
    if (_t->Pi[2] != NULL)
	phgMatDestroy(&_t->Pi[2]);
    if (_t->A_G != NULL)
	phgMatDestroy(&_t->A_G);
    if (_t->A_Pi != NULL)
	phgMatDestroy(&_t->A_Pi);

#if TEST_HIGH_ORDER
    if (_t->G0 != NULL)
	phgMatDestroy(&_t->G0);
    if (_t->Pi0[0] != NULL)
	phgMatDestroy(&_t->Pi0[0]);
    if (_t->Pi0[1] != NULL)
	phgMatDestroy(&_t->Pi0[1]);
    if (_t->Pi0[2] != NULL)
	phgMatDestroy(&_t->Pi0[2]);
    if (_t->A_G0 != NULL)
	phgMatDestroy(&_t->A_G0);
    if (_t->A_Pi0 != NULL)
	phgMatDestroy(&_t->A_Pi0);
#endif	/* TEST_HIGH_ORDER */

#if HYPRE_TEST
    if (_t->solver_hypre != NULL)
	phgSolverDestroy(&_t->solver_hypre);
#endif	/* HYPRE_TEST */
    if (_t->solver_G != NULL)
	phgSolverDestroy(&_t->solver_G);
    if (_t->solver_Pi != NULL)
	phgSolverDestroy(&_t->solver_Pi);

    phgFree(_t->cycle_type);
    phgFree(_t->aux_solver_opts);

    phgFree(solver->oem_data);
    solver->oem_data = NULL;

    return 0;
}

static int
Assemble(SOLVER *solver)
{
    int i, order;
    MAT_ROW *row;
    OEM_SOLVER *oem_solver;

    if (_t->assembled)
	return 0;
    _t->assembled = TRUE;

    assert(solver->mat->rmap->dofs != NULL);
    order = DofTypeOrder(solver->mat->rmap->dofs[0]->type);

    if (!_t->coefficients_defined) {
	/* use default coefficients alpha=1 and beta=1 */
	_t->coefficients_defined = TRUE;
	/* TODO: in this case construct and use G^t A G and Pi^t A Pi */
	/*phgSolverAMSSetConstantCoefficients(solver, 1.0, 1.0);*/
    }
#if TEST_HIGH_ORDER
    /* first order preconditioners */
    build_G_Pi(solver, 1, &_t->G,_t->Pi);
    build_poisson(solver, 1, _t->alpha,_t->beta, &_t->A_Pi,&_t->A_G);
    if (order > 1) {
	/* high order preconditioners */
	build_G_Pi(solver, order, &_t->G0,_t->Pi0);
	build_poisson(solver, order, _t->alpha,_t->beta, &_t->A_Pi0,&_t->A_G0);
    }
#else	/* TEST_HIGH_ORDER */
    /* preconditioners */
    build_G_Pi(solver, order, &_t->G,_t->Pi);
    build_poisson(solver, order, _t->alpha,_t->beta, &_t->A_Pi,&_t->A_G);
#endif	/* TEST_HIGH_ORDER */

    /* build aux solvers */
    phgOptionsPush();
    phgOptionsSetFloat("-solver_rtol", _t->aux_solver_rtol);
    phgOptionsSetInt("-solver_maxit", _t->aux_solver_maxit);
    oem_solver = phgSolverList[_t->aux_solver_index];
    if (phgVerbosity > 0)
	phgPrintf("    Auxiliary solver: %s\n", oem_solver->name);
    /* some default solver options */
    if (oem_solver == SOLVER_HYPRE && SOLVER_HYPRE != NULL) {
	phgOptionsSetOptions("-hypre_solver boomeramg -hypre_pc none");
	phgOptionsSetKeyword("-hypre_amg_coarsen_type", "hmis");
	phgOptionsSetInt("-hypre_amg_max_levels", 25);
	/* HYPRE_BoomerAMGSetAggNumLevels 1 (default) */
	phgOptionsSetKeyword("-hypre_amg_relax_type", "gs-h-forward");

	/* Gauss-Seidel instead of direct solver on coarsest grid */
	phgOptionsSetKeyword("-hypre_amg_coarsest_relax_type", "gs-h-forward");

	/* HYPRE_BoomerAMGSetNumSweeps 1 (done by solver-hypre.c) */
	phgOptionsSetFloat("-hypre_amg_strength", 0.25);	/* theta */
    }
    else if (oem_solver == SOLVER_TRILINOS && SOLVER_TRILINOS != NULL) {
	phgOptionsSetOptions("-aztec_solver cg -aztec_pc ml");
	phgOptionsSetOptions("-aztec_disable_errmsg");
#if HAVE_FEENABLEEXCEPT
	phgOptionsSetOptions("-aztec_disable_fpetrap");
#endif	/* HAVE_FEENABLEEXCEPT */
    }
    /* append user options */
    if (_t->aux_solver_opts != NULL)
	phgOptionsSetOptions(_t->aux_solver_opts);
    if (_t->A_G != NULL) {
	_t->solver_G = phgMat2Solver(oem_solver, _t->A_G);
	phgVecDestroy(&_t->solver_G->rhs);
	_t->solver_G->rhs_updated = TRUE;
    }
    _t->solver_Pi = phgMat2Solver(oem_solver, _t->A_Pi);
    phgVecDestroy(&_t->solver_Pi->rhs);
    _t->solver_Pi->rhs_updated = TRUE;
    phgOptionsPop();

#if HYPRE_TEST
    /* create a HYPRE AMS preconditioner for testing */
    phgOptionsPush();
    phgOptionsSetOptions("-hypre_solver ams -hypre_pc none");
    _t->solver_hypre = phgMat2Solver(SOLVER_HYPRE, solver->mat);
    phgVecDestroy(&_t->solver_hypre->rhs);
    _t->solver_hypre->rhs_updated = TRUE;
    phgSolverHypreAMSSetConstantPoisson(_t->solver_hypre, 1.0, 1.0);
    phgOptionsPop();
#endif	/* HYPRE_TEST */

    /* penalize boundary rows of A_G */
    for (i = 0; _t->A_G != NULL && i < _t->A_G->rmap->nlocal; i++) {
	row = _t->A_G->rows + i;
	if (row->ncols == 0) {
	    /* empty row */
	    row->ncols = row->alloc = 1;
	    row->cols = phgAlloc(sizeof(*row->cols));
	    row->cols[0] = i;
	    row->data = phgAlloc(sizeof(*row->data));
	    row->data[0] = DBL_MAX;
	}
	if (row->ncols == 1 && row->cols[0] == i)
	    row->data[0] = DBL_MAX;
    }

    /* penalize boundary rows of A_Pi */
    for (i = 0; i < _t->A_Pi->rmap->nlocal; i++) {
	row = _t->A_Pi->rows + i;
	assert(row->ncols > 0);
	if (row->ncols == 1 && row->cols[0] == i)
	    row->data[0] = DBL_MAX;
    }

    return 0;
}

VEC *
phgSolverGaussSeidel(MAT *mat, VEC *rhs, VEC **x_ptr,
		     INT maxit, FLOAT tol, FLOAT *res_ptr)
/* scaled symmetric hybrid Gauss-Seidel/Jacobi relaxation */
{
    VEC *x;
    INT it, i, j, n, *pc, *ncols;
    FLOAT *pd, a, b, res = DBL_MAX, *diag1;

    x = (x_ptr == NULL ? NULL : *x_ptr);
    if (x == NULL) {
	x = phgMapCreateVec(mat->cmap, rhs->nvec);
	if (x_ptr != NULL);
	    *x_ptr = x;
    }

    if (mat->type == PHG_UNPACKED)
	phgMatPack(mat);

    assert(mat->type == PHG_PACKED);

    if (mat->diag1 == NULL) {
	/* setup off-diagonal L1 norm1 */
	diag1 = mat->diag1 = phgAlloc(mat->rmap->nlocal * sizeof(*mat->diag1));
	if (mat->diag == NULL)
	    phgMatSetupDiagonal(mat);
	for (i = 0; i < mat->rmap->nlocal; i++) {
	    diag1[i] = Fabs(mat->diag[i]);
	    assert(diag1[i] != 0.);
	}
    }
    diag1 = mat->diag1;

    ncols = phgAlloc(x->map->nlocal * sizeof(*ncols));

    for (it = 0; it < maxit; it++) {
	/* forward scan */
	pc = mat->packed_cols;
	pd = mat->packed_data;
	for (i = 0; i < mat->rmap->nlocal; i++) {
	    a = rhs->data[i];

	    /* in-process columns */
	    ncols[i] = n = *(pc++);
	    for (j = 0; j < n; j++)
		a -= *(pd++) * x->data[*(pc++)];

	    x->data[i] += a / diag1[i];
	}

	/* backward scan */

	res = 0.0;
	for (i = mat->rmap->nlocal - 1; i >= 0; i--) {
	    a = rhs->data[i];

	    /* in-process columns */
	    n = ncols[i];
	    for (j = 0; j < n; j++)
		a -= *(--pd) * x->data[*(--pc)];
	    --pc;

	    b = x->data[i];
	    x->data[i] += a / diag1[i];
	    b = Fabs(x->data[i] - b);
	    if (res <= b)
		res = b;
	}

	if (phgVerbosity > 0)
	    phgPrintf("\tG-S sweep %d, res %le\n", it, (double)res);
	if (res <= tol)
	    break;
    }

    phgFree(ncols);

    if (res_ptr != NULL)
	*res_ptr = res;

    return x;
}

static void
mult_prec(SOLVER *solver, VEC *r, VEC *x, const char *cycle)
{
    int i;
    INT verbosity = phgVerbosity;
    VEC *r0 = NULL, *r1 = NULL, *x1 = NULL;

    while (*cycle != '\0') {
	if (*cycle < '0' && *cycle > '2')
	    phgError(1, "invalid AMS cycle type string: %s\n", _t->cycle_type);
	switch (*(cycle++) - '0') {
	    case 0:	/* smoothing */
		if (phgVerbosity > 1)
		    phgPrintf("\tAMS smoothing\n");
		phgVerbosity--;
		phgSolverGaussSeidel(solver->mat, r, &x, 1, 0., NULL);
		phgVerbosity = verbosity;
#if TEST_HIGH_ORDER
		if (_t->A_Pi0 != NULL) {
		    /* smoothing in (H1)^3 space */
		    phgVecCopy(r, &r0);
		    phgMatVec(MAT_OP_N, -1.0, solver->mat, x, 1.0, &r0);
		    phgVecDestroy(&r1);
		    phgVecDestroy(&x1);
		    x1 = phgMapCreateVec(_t->A_Pi0->cmap, 1);
		    for (i = 0; i < Dim; i++) {
			phgMatVec(MAT_OP_T, 1.0, _t->Pi0[i], r0, 0.0, &r1);
			phgVerbosity--;
			phgSolverGaussSeidel(_t->A_Pi0, r1, &x1, 1, 0.0, NULL);
			phgVerbosity = verbosity;
			phgMatVec(MAT_OP_N, 1.0, _t->Pi0[i], x1, 1.0, &x);
		    }
		    phgVecDestroy(&r1);
		    phgVecDestroy(&x1);
		}
		if (_t->A_G0 != NULL) {
		    /* smoothing in gradient space */
		    phgVecCopy(r, &r0);
		    phgMatVec(MAT_OP_N, -1.0, solver->mat, x, 1.0, &r0);
		    phgVecDestroy(&r1);
		    phgVecDestroy(&x1);
		    x1 = phgMapCreateVec(_t->A_Pi0->cmap, 1);
		    phgMatVec(MAT_OP_T, 1.0, _t->G0, r0, 0.0, &r1);
		    phgVerbosity--;
		    phgSolverGaussSeidel(_t->A_G0, r1, &x1, 1, 0.0, NULL);
		    phgVerbosity = verbosity;
		    phgMatVec(MAT_OP_N, 1.0, _t->G0, x1, 1.0, &x);
		    phgVecDestroy(&r1);
		    phgVecDestroy(&x1);
		}
#endif	/* TEST_HIGH_ORDER */
		break;
	    case 1:	/* space G correction */
		if (_t->solver_G == NULL)
		    break;
		if (phgVerbosity > 1)
		    phgPrintf("\tAMS space G correction\n");
		phgVecCopy(r, &r0);
		phgMatVec(MAT_OP_N, -1.0, solver->mat, x, 1.0, &r0);
		if (x1 == NULL)
		    x1 = phgMapCreateVec(_t->G->cmap, 1);
		phgMatVec(MAT_OP_T, 1.0, _t->G, r0, 0.0, &r1);
		_t->solver_G->rhs = r1;
		_t->solver_G->rhs->assembled = TRUE;
		bzero(x1->data, x1->map->nlocal * sizeof(*x1->data));
		phgVerbosity--;
		phgSolverVecSolve(_t->solver_G, FALSE, x1);
		phgVerbosity = verbosity;
		_t->solver_G->rhs = NULL;
		phgMatVec(MAT_OP_N, 1.0, _t->G, x1, 1.0, &x);
		break;
	    case 2:	/* space Pi correction */
		if (phgVerbosity > 1)
		    phgPrintf("\tAMS space Pi correction\n");
		phgVecCopy(r, &r0);
		phgMatVec(MAT_OP_N, -1.0, solver->mat, x, 1.0, &r0);
		if (x1 == NULL)
		    x1 = phgMapCreateVec(_t->Pi[0]->cmap, 1);
		phgVerbosity--;
		for (i = 0; i < Dim; i++) {
#if 0
		    if (i > 0) {
			/* multiplicative corrections of the three components */
			phgVecCopy(r, &r0);
			phgMatVec(MAT_OP_N, -1.0, solver->mat, x, 1.0, &r0);
		    }
#endif
		    phgMatVec(MAT_OP_T, 1.0, _t->Pi[i], r0, 0.0, &r1);
		    _t->solver_Pi->rhs = r1;
		    _t->solver_Pi->rhs->assembled = TRUE;
		    bzero(x1->data, x1->map->nlocal * sizeof(*x1->data));
		    phgSolverVecSolve(_t->solver_Pi, FALSE, x1);
		    _t->solver_Pi->rhs = NULL;
		    phgMatVec(MAT_OP_N, 1.0, _t->Pi[i], x1, 1.0, &x);
		}
		phgVerbosity = verbosity;
		break;
	}	/* case */
    }	/* while */

    if (r0 != NULL)
	phgVecDestroy(&r0);
    if (r1 != NULL)
	phgVecDestroy(&r1);
    if (x1 != NULL)
	phgVecDestroy(&x1);
}

static int
Solve(SOLVER *solver, VEC *x, BOOLEAN destroy)
{
    char *p = _t->cycle_type, *q;
    VEC *y0 = NULL, *x0 = NULL;
    FLOAT res = DBL_MAX, res_b = DBL_MAX, b_norm = 0.0;

    Assemble(solver);

    if (phgVerbosity > 0)
	phgPrintf("    AMS cycle type: %s\n", _t->cycle_type);

    if (solver->maxit > 0 && (solver->rtol > 0. || solver->atol > 0.)) {
	b_norm = phgVecNorm2(solver->rhs, 0, NULL);
	if (b_norm == 0.0) {
	    solver->nits = 0;
	    solver->residual = 0.;
	    return 0;
	}
    }

    solver->nits = 0;
    while (solver->nits < solver->maxit) {
	solver->nits++;
#if HYPRE_TEST
	/* test preconditioning using Hypre AMS */
	_t->solver_hypre->rtol = solver->rtol;
	_t->solver_hypre->maxit = solver->maxit;
	_t->solver_hypre->rhs = solver->rhs;
	phgSolverVecSolve(_t->solver_hypre, FALSE, x);
	_t->solver_hypre->rhs = NULL;
	solver->residual = _t->solver_hypre->residual;
	return solver->nits = _t->solver_hypre->nits;
#endif	/* HYPRE_TEST */

	if ((q = strchr(p, '+')) == NULL) {
	    mult_prec(solver, solver->rhs, x, p);
	}
	else {
	    phgVecCopy(solver->rhs, &y0);
	    phgMatVec(MAT_OP_N, -1.0, solver->mat, x, 1.0, &y0);
	    x0 = phgMapCreateVec(y0->map, y0->nvec);
	    while (*p != '\0') {
		if (q != NULL)
		    *q = '\0';
		bzero(x0->data, sizeof(*x0->data) * x0->map->nlocal);
		mult_prec(solver, y0, x0, p);
		phgVecAXPBY(1.0, x0, 1.0, &x);
		if (q == NULL)
		    break;
		*q = '+';
		q = strchr(p = q + 1, '+');
	    }
	}

	if (solver->rtol > 0.0 || solver->atol > 0.0 || phgVerbosity > 0) {
	    phgVecCopy(solver->rhs, &x0);
	    phgMatVec(MAT_OP_N, -1.0, solver->mat, x, 1.0, &x0);
	    res = phgVecNorm2(x0, 0, NULL);
	    res_b = res / b_norm;
	    if (phgVerbosity > 0)
		phgPrintf("    AMS: it = %d, residual = %le, res_b = %le\n",
				solver->nits, (double)res, (double)res_b);
	    if (res <= solver->atol || res_b <= solver->rtol)
		break;
	}
    }

    phgVecDestroy(&x0);
    phgVecDestroy(&y0);

    if (destroy) {
	Destroy(solver);
	phgMatDestroy(&solver->mat);
    }

    solver->residual = res/*_b*/;

    return solver->nits;
}

OEM_SOLVER phgSolverAMS_ = {
    "AMS", RegisterOptions,
    Initialize, Finalize, Init, Create, Destroy, AddMatrixEntries,
    AddRHSEntries, Assemble, Solve
};
