/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2007 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: solver-spc.c,v 1.24 2007/09/16 05:50:21 zlb Exp $ */

/* The SPC interface for PHG */

#include "phg.h"

#if USE_SPC

#include <spc.h>
#include <stdlib.h>

typedef struct {
    vec		B;	/* B=RHS */
    vec		U;	/* U=solution */
    mat		A;	/* A=matrix */
    map		MAP;
    pc		PC;
    solver	SOLVER;
    BOOLEAN	assembly_rhs;
    BOOLEAN	assembly_mat;

    int		solver_index;
    int		pctype;
    int		pcilutype;
    INT		solver_ksize;
    FLOAT	pc_tol;
    INT		pc_fill;
    INT		pc_overlap;
} OEM_DATA;

/* convenience macro */
#define _t	((OEM_DATA *)solver->oem_data)

#define Initialize	NULL
#define Finalize	NULL

/* Options */

static const char *solver_names[] = {"gmres", "bicgstab", NULL};
static const SOLVERTYPE solver_list[] = {SOLGMRES, SOLBICGSTAB};
static int solver_index = 0;

static const char *pctype_names[] = {"bj", "ras", "schras", "mschras", NULL};
static const PCTYPE pctype_list[] = {PCBJ, PCRAS, PCSCHRAS, PCMSCHRAS};
static int pctype = 2;

static const char *pcilutype_names[] = {"ilu0", "iluk", "ilut", "ilubst",
					"iluc", "ilutc", "arms", NULL};
static const char pcilutype_list[] = {PCILU0, PCILUK, PCILUT, PCILUBST,
					PCILUC, PCILUTC, PCARMS};
static int pcilutype = 2;

static INT	solver_ksize = 30;
static FLOAT	pc_tol = 0.01;
static INT	pc_fill = 50;
static INT	pc_overlap = 1;

static int
RegisterOptions(void)
{
    phgOptionsRegisterTitle("\nSPC options:", NULL, "spc");

    /* Options for solver */
    phgOptionsRegisterKeyword("spc_solver_type", "Solver type",
				solver_names, &solver_index);
    phgOptionsRegisterInt("spc_solver_ksize", "Krylov subspace size "
			  "(restart size)", &solver_ksize);

    phgOptionsRegisterTitle("", NULL, NULL);

    /* Options for preconditioner */
    phgOptionsRegisterKeyword("spc_pc_type", "Preconditioner type",
				pctype_names, &pctype);
    phgOptionsRegisterKeyword("spc_pc_local_type", "Local preconditioner type",
				pcilutype_names, &pcilutype);
    phgOptionsRegisterFloat("spc_pc_tol", "Drop tolerance for ILUT", &pc_tol);
    phgOptionsRegisterInt("spc_pc_fill", "Fill-in parameter", &pc_fill);
    phgOptionsRegisterInt("spc_pc_overlap", "Overlapping level", &pc_overlap);

    return 0;
}

static int
Init(SOLVER *solver)
{
    solver->oem_data = phgAlloc(sizeof(OEM_DATA));

    _t->solver_index	= solver_index;
    _t->pctype		= pctype;
    _t->pcilutype	= pcilutype;
    _t->solver_ksize	= solver_ksize;
    _t->pc_tol		= pc_tol;
    _t->pc_fill		= pc_fill;
    _t->pc_overlap	= pc_overlap;

    return 0;
}

static int
Create(SOLVER *solver)
{
    MAP *map = solver->mat->rmap;
    int i, j, N = map->nglobal, *part;

    assert(solver->mat->type != PHG_MATRIX_FREE);
    assert(solver->mat->assembled);	/* only works with assembled matrix */

    _t->assembly_rhs = FALSE;
    _t->assembly_mat = FALSE;
    
    /* create map */
    part = phgAlloc(N * sizeof(*part));
#if 1 
    for (i = 0, j = 1; i < N; i++) {
	while (i >= solver->partition[j]) {
	    if (++j > map->nprocs)
		phgError(1, "%s:%d, unexpected error!\n", __FILE__, __LINE__);
	}
	part[i] = j - 1;
    }
#else 
    for (i = 0; i < N; i++) {
       part[i] = 0;
    }
#endif

    CreateMapMetis(&_t->MAP, N, part, map->comm, 0, 1, NONINTERLACED);
    phgFree(part);

    /* create matrix */
    CreateMat(&_t->A, _t->MAP);

    /* create vectors */
    CreateVec(&_t->B, _t->MAP);
    CreateVec(&_t->U, _t->MAP);
    /*VecPut(_t->U, 1.0);*/

    return 0;
}

static int
Destroy(SOLVER *solver)
{
    if (solver->oem_data == NULL)
	return 0;

    if (_t->A != NULL)
	DeleteMat(&_t->A);
    if (_t->B != NULL)
	DeleteVec(&_t->B);
    if (_t->U != NULL)
	DeleteVec(&_t->U);
    DeleteMap(&_t->MAP);

    phgFree(solver->oem_data);
    solver->oem_data = NULL;

    return 0;
}

static int
AddMatrixEntries(SOLVER *solver, INT nrows, INT *rows, INT ncols, INT *cols,
			FLOAT *values)
{
    if (sizeof(INT)!=sizeof(int) || sizeof(FLOAT)!=sizeof(double))
	phgError(1, "%s:%d: unimplemented.\n", __FILE__, __LINE__);
    MatSetValues(_t->A, nrows, rows, ncols, cols, values, INSERT_VALUES);
    _t->assembly_mat = TRUE;

    return 0;
}

static int
AddRHSEntries(SOLVER *solver, INT n, INT *ni, FLOAT *values)
{
    if (sizeof(INT)!=sizeof(int) || sizeof(FLOAT)!=sizeof(double))
	phgError(1, "%s:%d: unimplemented.\n", __FILE__, __LINE__);
    VecSetValues(_t->B, n, ni, values, INSERT_VALUES);
    _t->assembly_rhs = TRUE;

    return 0;
}
 
static int
Assemble(SOLVER *solver)
{
    if (_t->assembly_rhs) {
	VecSetup(_t->B);
	_t->assembly_rhs = FALSE;
    }

    if (_t->assembly_mat) {
	MatSetup(_t->A);
	_t->assembly_mat = FALSE;
    }
    
    return 0;
}

static int
set_initial_solution(SOLVER *solver, VEC *x)
{
    double *vec;
    int size;
    INT i;

    size = VecGetLocalSize(_t->U);
    VecGetArray(_t->U, &vec);
    for (i = 0; i < size; i++)
	vec[i] = x->data[i];
    VecRestoreArray(_t->U, &vec);

    return 0;
}

static int
copy_solution(SOLVER *solver, VEC *x)
/* copy solution from SPC's vec 'U' to local DOFs */
{
    double *vec;
    int size;
    INT i;

    size = VecGetLocalSize(_t->U);
    VecGetArray(_t->U, &vec);
    for (i = 0; i < size; i++)
	x->data[i] = vec[i];
    VecRestoreArray(_t->U, &vec);

    return 0;
}

static int
Solve(SOLVER *solver, VEC *x, BOOLEAN destroy)
{
    int nits, ret;
    char s[32];

    Assemble(solver);

    set_initial_solution(solver, x);

#if 0
    {
	double norma = 0.0, normb, normu;
	/*MatNorm(_t->A, NORM_FROBENIUS, &norma);*/
	VecGetNorm2(_t->B, &normb);
	VecGetNorm2(_t->U, &normu);
	phgPrintf("\nNorm(A) = %lg, Norm(b) = %lg, Norm(x0) = %lg\n",
			(double)norma, (double)normb, (double)normu);
    }
#endif

    /* create PC */
    CreatePC(&_t->PC, _t->A);
    /* Type - PCSCHRAS, PCRAS, PCBJ */
    PCSetType	  (_t->PC, pctype_list[_t->pctype]);
    PCSetLocalType(_t->PC, pcilutype_list[_t->pcilutype]);
    PCSetTol	  (_t->PC, _t->pc_tol);
    PCSetFill	  (_t->PC, _t->pc_fill);
    PCSetOvp	  (_t->PC, _t->pc_overlap);
    PCSetup	  (_t->PC);

    /* create solver */
    CreateSolver(&_t->SOLVER, _t->A, _t->PC);
    SolverSetType (_t->SOLVER, solver_list[_t->solver_index]);

    sprintf(s, "%d", solver->maxit);
    SolverSetParam(_t->SOLVER, PARAMMAXITS, s);

    sprintf(s, "%d", _t->solver_ksize);
    SolverSetParam(_t->SOLVER, PARAMKSIZE, s);

    sprintf(s, "%0.16le", (double)solver->rtol);
    SolverSetParam(_t->SOLVER, PARAMTOL, s);

    ret = SolverApply(_t->SOLVER, _t->B, _t->U);
    nits = SolverGetIts(_t->SOLVER);
    if (ret != 0)
	nits = (nits == 0) ? -1 : ((nits < 0) ? nits : -nits);
    solver->residual = 0.0;	/* fake residual */

    DeleteSolver(&_t->SOLVER);
    DeletePC(&_t->PC);

    if (destroy) {
	DeleteMat(&_t->A);
	_t->A = NULL;
	DeleteVec(&_t->B);
	_t->B = NULL;
    }

    copy_solution(solver, x);

    if (destroy) {
	DeleteVec(&_t->U);
	_t->U = NULL;
    }

    return nits;
}

/*--------------------------------------------------------------------*/

OEM_SOLVER phgSolverSPC_ = {
    "SPC", RegisterOptions,
    Initialize, Finalize, Init, Create, Destroy, AddMatrixEntries,
    AddRHSEntries, Assemble, Solve
};

#endif		/* USE_SPC */
